import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)

import java.lang.Math;
import java.util.*;
import java.awt.Color;

public class Explosion extends Game {
    private static final GreenfootImage[][] explosionStates = new GreenfootImage[7][4];
    private final static Random RAND = new Random();

    // Requires zelda_sprite
    public static void loadFrames(GreenfootImageCutter sprite) {

        // Middle
        explosionStates[0] = sprite.getSubImgs(3, 159, 2, 20, 20, 4);
        // Left
        explosionStates[1] = sprite.getSubImgsVariableWidth(91, 159, 2, new int[] { 18, 19, 20, 20 }, 20);

        // Top and bottom
        for (int i = 0; i < 2; i++) {
            explosionStates[2 + i] = sprite.getSubImgsVariableWidth(3, 184 + i * 20, 2, new int[] { 11, 15, 20, 20 }, 20);
        }

        // Right
        explosionStates[4] = sprite.getSubImgsVariableWidth(3, 229, 2, new int[] { 18, 19, 20, 20 }, 20);
        // Horizontal
        explosionStates[5] = sprite.getSubImgs(77, 204, 2, 20, 20, 4);

        // Vertical
        explosionStates[6] = sprite.getSubImgsVariableWidth(88, 229, 2, new int[] { 12, 15, 20, 20 }, 18);

        for (int x = 0; x < explosionStates.length; x++) {
            for (int y = 0; y < 4; y++) {
                GreenfootImage current = explosionStates[x][y];
                if (current == null) continue;
                double factor = 32. / (ExplosionPart.values()[x].isVerticalElement() ? current.getHeight() : current.getWidth());
                current.scale((int) (factor * current.getWidth()), (int) (factor * current.getHeight()));
                ImageUtils.replaceColor(current, new Color(78, 230, 78), new Color(0, 0, 0, 0));
            }
        }

    }

    private int state, maxRange;
    private List<SubExplosion> parts = new ArrayList<>();
    private Hitbox box = new Hitbox(32, 32, this);

    // maxRange = player attribute, upgradable using items
    public Explosion(int maxRange) {
        this.maxRange = maxRange;
        setImage(new GreenfootImage(1, 1));
    }

    private void buildExplosionParts(Direction dir, int movement) {
        int modX = dir.getModX() * 32;
        int modY = dir.getModY() * 32;

        int currentX = getX();
        int currentY = getY();

        for (int i = 0; i < movement; i++) {
            currentX += modX;
            currentY += modY;

            SubExplosion sub = new SubExplosion(i == movement - 1 ? ExplosionPart.getEndPart(dir) : ExplosionPart.getTransitoryPart(dir));
            getWorld().addObject(sub, currentX, currentY);
            parts.add(sub);
        }

    }

    @Override
    public void firstAct() {
        setLocation(getFieldCenterX(), getFieldCenterY());

        getWorld().addObject(box, getX(), getY());

        SubExplosion sub = new SubExplosion(ExplosionPart.CENTER);
        getWorld().addObject(sub, getX(), getY());
        parts.add(sub);

        for (Direction direction : Direction.values()) {
            int movement = 0;

            int modX = direction.getModX() * 32;
            int modY = direction.getModY() * 32;

            for (int i = 0; i < maxRange; i++) {
                box.move(modX, modY);
                Set<Class<? extends Actor>> touching = box.getIntersectingTypes();
                if (touching.contains(Block.class)) {
                    break;
                } else if (touching.contains(DestroyableBlock.class)) {
                    movement++;
                    break;
                }
                movement++;
            }
            box.setLocation(getX(), getY());
            buildExplosionParts(direction, movement);
        }
    }

    @Override
    public void doAct() {
        // every 7 ticks
        if (++ticksAlive % 7 == 0) {
            if (state++ >= 3) {
                Set<Actor> touching = new HashSet<>();
                parts.stream().map(SubExplosion::getTouching).forEach(touching::addAll);

                for (Actor actor : touching) {
                    if (actor instanceof DestroyableBlock) {
                        if (RAND.nextInt(100) < 32)
                            getWorld().addObject(new Item(Resources.item_names.get(RAND.nextInt(Resources.item_names.size()))), actor.getX(), actor.getY());
                        getWorld().removeObject(actor);
                    } else if (actor instanceof Hitbox) {
                        Actor source = ((Hitbox) actor).getSource();

                        if (source instanceof LocalPlayer)
                            ((LocalPlayer) source).checkDeath();
                        else if (source instanceof Bomb)
                            ((Bomb) source).timeToExplosion = 0;
                            
                    } else if (actor instanceof Item) {
                        getWorld().removeObject(actor);
                    }
                }
                
                parts.forEach(getWorld()::removeObject);
                getWorld().removeObject(box);
                getWorld().removeObject(this);
                return;
            }
            parts.forEach(SubExplosion::updateState);
        }
    }

    public class SubExplosion extends Game {

        private ExplosionPart part;
        private SubExplosion(ExplosionPart part) {
            this.part = part;

            updateState();
        }

        public void updateState() {
            setImage(part.getImage(state % 4));
        }

        public List<Actor> getTouching() {
            return getIntersectingObjects(null);
        }

    }

    private enum ExplosionPart {
        CENTER      (false),
        LEFT        (false),
        TOP         (true),
        BOTTOM      (true),
        RIGHT       (false),
        HORIZONTAL  (false),
        VERTICAL    (true);

        public static ExplosionPart getEndPart(Direction dir) {
            switch(dir) {
                case RIGHT:     return RIGHT;
                case LEFT:      return LEFT;
                case UP:        return TOP;
                case DOWN:      return BOTTOM;
            }

            return null;
        }

        public static ExplosionPart getTransitoryPart(Direction dir) {
            return dir == Direction.RIGHT || dir == Direction.LEFT ? HORIZONTAL : VERTICAL;
        }

        private boolean vertical;

        ExplosionPart(boolean vertical) {
            this.vertical = vertical;
        }

        public GreenfootImage getImage(int state) {
            return explosionStates[this.ordinal()][state];
        }

        public boolean isVerticalElement() {
            return this.vertical;
        }
    }
}
